import os
from pathlib import Path
import lark
from dicey.dietransformer import DieTransformer


class DieParser:
    def __init__(self):
        self.grammar = open(
            os.path.join(Path(__file__).parent, "dicey.lark")
        ).read()
        self.last_exp = ""
        self.intermediate_expr = []
        self.results = []
        self.transformer = DieTransformer()
        self.parser = lark.Lark(
            self.grammar,
            parser="lalr",
            transformer=self.transformer,
            propagate_positions=False,
        )

    def parse(self, text):
        self.__init__()
        self.parser.parse(text)
        self.transformer._eval()
        self.last_exp = self.transformer.string
        self.intermediate_expr.append(self.transformer.intermediate_expr)
        self.results.append(self.transformer.value.get())

        for i in range(self.transformer.repeats - 1):
            self.transformer.__init__()
            self.parser.parse(text)
            self.transformer._eval()
            self.intermediate_expr.append(self.transformer.intermediate_expr)
            self.results.append(self.transformer.value.get())

    def __str__(self):
        s = ""
        if len(self.results) > 1:
            s += "{} = ".format(self.last_exp)
            s += "\n{}".format(
                "-"
                * (
                    len(self.intermediate_expr[0])
                    + 3
                    + len(str(max(self.results)))
                )
            )
            for i in range(self.transformer.repeats):
                s += "\n{} = {}".format(
                    self.intermediate_expr[i], self.results[i]
                )
        else:
            s += "{} = ".format(self.last_exp)
            s += "{} = {}".format(self.intermediate_expr[0], self.results[0])
        return s



from lark import Lark
import random

try:
    from dicey import result
except ImportError:
    import result

roll_grammar = """
    start: expr

    expr: token+

    token: op* (die|mod)

    die: numdie "d" max

    !op: "-"|"+"

    numdie: INT
    max: INT
    mod: INT

    %import common.INT
    %import common.WS
    %ignore WS
"""


class DieRoller:
    def __init__(self):
        self.parser = Lark(roll_grammar, parser="lalr")
        self.grammar = roll_grammar

        self.input_expr = ""
        self.die_expr = ""
        self.roll_expr = ""
        self.mod_expr = ""
        self.total = 0

        self.result = result.DieRollResult()

    def roll(self, expr):
        self.input_expr = expr
        self.parse_tree = self.parser.parse(expr)

        (
            self.result.die_expr,
            self.result.roll_expr,
            self.result.mod_expr,
            self.result.total,
        ) = self.process_tree(self.parse_tree)

    def reroll(self):
        self.roll(self.input_expr)

    def process_tree(self, t):

        positive = True
        total = 0
        die_expr = ""
        roll_expr = ""
        mod_expr = ""

        for node in t.find_data("token"):
            ntype = node.children[0].data

            if ntype == "die":
                die, roll, result = self._process_die(node.children[0])
                roll_expr += "{0}".format(roll)
                die_expr += "{0}".format(die)
                total += result
            elif ntype == "mod":
                value = int(node.children[0].children[0])
                if value < 0:
                    mod_expr += " - {0}".format(value)
                    total -= value
                else:
                    mod_expr += " + {0}".format(value)
                    total += value
            else:
                ttype = node.children[1].data
                positive = node.children[0].children[0] == "+"

                if ttype == "mod":
                    value = int(node.children[1].children[0])
                    if positive:
                        mod_expr += " + {0}".format(value)
                        total += value
                    else:
                        mod_expr += " - {0}".format(value)
                        total -= value
                else:
                    die, roll, result = self._process_die(node.children[1])
                    if positive:
                        if die_expr == "":
                            roll_expr += "{0}".format(roll)
                            die_expr += "{0}".format(die)
                        else:
                            roll_expr += " + {0}".format(roll)
                            die_expr += " + {0}".format(die)
                        total += result
                    else:
                        roll_expr += " - [{0}]".format(roll)
                        die_expr += " - {0}".format(die)
                        total -= result

        return (die_expr, roll_expr, mod_expr, total)

    def _process_die(self, t):
        numdie = int(t.children[0].children[0])
        maxdie = int(t.children[1].children[0])

        die_expr = "{0}d{1}".format(numdie, maxdie)
        roll_expr = ""
        total = 0
        for die in range(numdie):
            roll = random.randint(1, maxdie)
            roll_expr += "({0}) ".format(roll)
            total += roll
        roll_expr = roll_expr.replace(") (", ") + (")
        roll_expr = roll_expr[:-1]

        return (die_expr, roll_expr, total)



class DieRollResult:
    def __init__(self, die_expr="", roll_expr="", mod_expr="", total=0):

        self.die_expr = die_expr
        self.roll_expr = roll_expr
        self.mod_expr = mod_expr
        self.total = total

    def _check_mod(self):
        if self.die_expr == "" and self.mod_expr[0:3] == " + ":
            self.mod_expr = self.mod_expr[3:]

    def __str__(self):
        return self._total()

    def _total(self):
        """
        print result -> int
        """
        self._check_mod()
        return "{0}".format(self.total)

    def v(self):
        """
        print verbose result -> '1d20 + 2d4 + 1 = 16'
        """
        self._check_mod()
        return "{0}{1} = {2}".format(self.die_expr, self.mod_expr, self.total)

    def vv(self):
        """
        print very verbose result -> '1d20 + 2d4 + 1 = 10 + 3 + 2 + 1 = 16'
        """
        self._check_mod()
        return "{0}{1} = {2}{3} = {4}".format(
            self.die_expr,
            self.mod_expr,
            self.roll_expr,
            self.mod_expr,
            self.total,
        )



import random
import sys
from queue import LifoQueue

import lark


class DieTransformer(lark.Transformer):
    def __init__(self):
        self.string = ""
        self.intermediate_expr = ""
        self.value = LifoQueue()
        self.operators = LifoQueue()
        self.max = -(sys.maxsize)
        self.min = sys.maxsize
        self.repeats = 0

    def _precedence(self, op):
        if op == "+" or op == "-":
            return 1
        if op == "*" or op == "/":
            return 2
        return 0

    def _applyOp(self, a, b, op):
        if op == "+":
            return a + b
        if op == "-":
            return a - b
        if op == "*":
            return a * b
        if op == "/":
            return a / b

    def _eval(self):
        while not self.operators.empty():
            op = self.operators.get()
            b = self.value.get()
            a = self.value.get()
            self.value.put(self._applyOp(a, b, op))

    def lparen(self, args):
        self.string += "("
        self.intermediate_expr += "("

        self.operators.put(args[0])

    def rparen(self, args):
        self.string += ")"
        self.intermediate_expr += ")"

        op = self.operators.get()
        while op != "(":
            b = self.value.get()
            a = self.value.get()

            if op == "+":
                self.value.put(a + b)
            elif op == "-":
                self.value.put(a - b)
            elif op == "*":
                self.value.put(a * b)
            elif op == "/":
                self.value.put(a / b)

            op = self.operators.get()

    def num(self, args):
        self.string += "{}".format(args[0])
        self.intermediate_expr += args[0]

        self.value.put(int(args[0]))

    def op(self, args):
        self.string += " {} ".format(args[0])
        self.intermediate_expr += " {} ".format(args[0])

        if not self.operators.empty():
            op = self.operators.queue[-1]
            while not self.operators.empty() and self._precedence(
                op
            ) >= self._precedence(args[0]):
                op = self.operators.get()
                b = self.value.get()
                a = self.value.get()
                self.value.put(self._applyOp(a, b, op))

        self.operators.put(args[0])

    def die(self, args):
        self.string += "{}d{}".format(args[0], args[1])

        n = int(args[0])
        m = int(args[1])
        tmp_str = ""
        total = 0
        for die in range(n):
            roll = random.randrange(1, m + 1)
            total += roll
            tmp_str += "({})".format(roll)

            if roll > self.max:
                self.max = roll

            if roll < self.min:
                self.min = roll

        tmp_str = ") + (".join(tmp_str.split(")("))
        self.intermediate_expr += "{}".format(tmp_str)
        self.value.put(total)

    def meta(self, args):
        op = args[0]
        t = args[1]

        if t == "min":
            self.op([op])
            self.value.put(self.min)
            self.intermediate_expr += "[{}]".format(self.min)
        elif t == "max":
            self.op([op])
            self.value.put(self.max)
            self.intermediate_expr += "[{}]".format(self.max)

        self.string += "{}".format(args[1])

    def repeat(self, args):
        self.repeats = int(args[0])
        self.string += " {{{0}}}".format(self.repeats)
